"""Optimizers link the entire QD algorithm together.

.. autosummary::
    :toctree:

    ribs.optimizers.Optimizer
"""
from ribs.optimizers._optimizer import Optimizer

__all__ = [
    "Optimizer",
]
